/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 543 $ $LastChangedDate: 2002-08-25 16:28:24 +0200 (Sun, 25 Aug 2002) $
 *
 * Parser - Bison parser for Gont.
 *
 * Ast::* values are made using function from Cast module. This file is endless
 * source of seg-faults as C compiler cannot typecheck gont values...
 */
%{

#include "parser.h"
#include "support.h"

#include <gont/runtime.h>
#include "cgont/ast.h"
#include "cgont/cast.h"
#include "cgont/list.h"
#include "cgont/std.h"

#define YYPARSE_PARAM parm
#define YYLEX_PARAM parm
#define YYERROR_VERBOSE 1

#define YYDEBUG 1

/* Constants. */
static struct gont_union *none;
static struct gont_union *nil;
static struct gont_string *empty_string;

/* Used by impl: rule and parse routine. */
static struct gont_union *decls;

/* Defined in support.c. */
extern struct gont_string *input_filename;
struct GontS_Ast__location *make_location(YYLTYPE *l);
void yyerror(const char *msg);

extern YYLTYPE last_lex_loc;

#define make_expr(raw, l) Cast$$make_expr(raw, make_location(l))
#define make_tyvar Cast$$make_tyvar
#define make_located(d, l) Cast$$make_located(d, make_location(l))
#define make_ref_record(x) Cast$$make_ref_record(x)
#define make_vardef(t, n) Cast$$make_vardef(t, n)
#define make_field(t, n) Cast$$make_field(t, n, make_location(&yyloc))
#define make_tydef(name, vars, ty) \
	Cast$$make_tydef(name, vars, ty, make_location(&yyloc))
#define make_fn_decl(header, exportas, body) \
	Cast$$make_fn_decl(header, exportas, body)
#define bare_fn_type(r, a) Cast$$bare_fn_type(r, a)
#define make_vardefs(t, vars) Cast$$make_vardefs(t, vars)
#define make_lambda Cast$$make_lambda
#define make_lambda_assign Cast$$make_lambda_assign
#define make_break Cast$$make_break
#define make_loop Cast$$make_loop
#define make_namedexpr Cast$$make_namedexpr
#define make_sini_expr Cast$$make_sini_expr
#define make_named_type Cast$$make_named_type
#define make_try_record Cast$$make_try_record
#define make_float_const Cast$$make_float_const
#define make_int_const Cast$$make_int_const
#define make_named_pattern Cast$$make_named_pattern
#define make_section_record Cast$$make_section_record
#define make_case_clause Cast$$make_case_clause

#define U(fld, val) \
	gont_make_union(GontUF_Ast__ ## fld, (val))

#define Some(x) \
	gont_make_union(GontUF_Std__Some, (x))

#define P(a, b) \
	gont_make_tuple(2, (a), (b))

#define TRIPLE(a, b, c) \
	gont_make_tuple(3, (a), (b), (c))

#define CONS(a, b) \
	gont_make_union(GontUF_List__Cons, P((a), (b)))

#define EXPR(fld, val) \
	make_expr(U(fld, (val)), &yyloc)

#define BIN(op, e1, e2) \
	EXPR(SimpleBinOp, TRIPLE(U(op, NULL), (e1), (e2)))

#define UNARY(op, e) \
	EXPR(SimpleUnOp, P(U(op, none), e))

#define LOCATED(x) \
	make_located(x, &yyloc)

#define STMT(fld, val) \
	LOCATED(U(fld, (val)))

#define PAIR(a, b) \
	STMT(Pair, P((a), (b)))

#define SKIP() \
	STMT(Skip, NULL)

#define LOOP(lab, body, inc) \
	LOCATED(make_loop(lab, body, inc))
%}

/* parametrized with string. */
%token INT
%token FLOAT
%token ID
%token STRING
%token TYVAR	       /* 'a -> TYVAR("a") */

/* parametrized with int. */
%token CHAR		/* not used yet. */

/* Non parametrized. */
%token KW_INT KW_RETURN KW_VOID KW_SKIP KW_BOOL KW_FLOAT KW_IN
%token KW_EXTERN_C KW_STRING KW_FUN KW_IF KW_ELSE KW_WHILE KW_DO KW_FOR
%token KW_CASE KW_SWITCH KW_LET KW_EXPORT_AS KW_STRUCT KW_OPT_STRUCT
%token KW_UNION KW_TYPE KW_TYPEDEF KW_TRUE KW_FALSE KW_OPEN
%token KW_NULL KW_RAISE KW_TRY KW_EXCEPTION KW_BREAK KW_CONTINUE
%token KW_FINALLY KW_WITH KW_SECTION
%token DOUBLECOLON ARROW
%token GE LE EQ NE LTLT GTGT
%token PLUS_PLUS MINUS_MINUS
%token SLASH_EQUAL PLUS_EQUAL MINUS_EQUAL PERCENT_EQUAL STAR_EQUAL
%token AND_AND OR_OR

%pure_parser
%locations
%expect 1

/* lowest precedence */
%right '='
%left AND_AND OR_OR
%left '<' '>' GE LE
%left EQ NE
%left '+' '-' '@'
%left '*' '%' '/'
%nonassoc UMINUS
%right PLUS_PLUS MINUS_MINUS '!'
%nonassoc APPLICATION
%left '.'
/* highest precedence */

%%

/***********************************************************************
 * Compilation unit.
 ***********************************************************************/

module:
	  /* empty */		{ $$ = nil; decls = $$; }
	| decl module		{ $$ = CONS($1, $2); decls = $$; }
;

decl:
	nl_decl		{ $$ = LOCATED($1); }
;

nl_decl:
	  func
	| global_var
	| extern_c
	| type_decl
	| exn_decl
	| func_proto
	| top_open
	| dummy_decl
	| section
;


/***********************************************************************
 * Toplevel symbols.
 ***********************************************************************/
dummy_decl:
	';'			{ $$ = U(DummyDecl, NULL); }
;

top_open:
	KW_OPEN ID				{ $$ = U(TopLevelOpen, $2); }
;

global_var:
       vardef ';'			{ $$ = U(GlobalDecl, $1); }
;

exn_decl:
	KW_EXCEPTION vardef			{ $$ = U(ExceptionDecl, $2); }
;

string:
	  STRING			{ $$ = $1; }
	| STRING string			{ $$ = gont_strjoin($1, $2); }
;

maybe_export_as:
	 /* empty */				{ $$ = empty_string; }
       | KW_EXPORT_AS string			{ $$ = $2; }
;

func:
       func_header maybe_export_as fnc_body
	{ $$ = make_fn_decl($1, $2, $3); }

;

section:
	KW_SECTION ID fnc_body		
	{ $$ = U(SectionDecl, P($2, make_section_record($3))); }
;

/* from now on ne_ == non_empty_ */

extern_c:
	  KW_EXTERN_C func_header ';'		{ $$ = U(ExternC, $2); }
	| KW_EXTERN_C vardef ';'	{ $$ = U(ExternC, $2); }
;

func_def:
	func_header fnc_body
	{ $$ = LOCATED(make_lambda_assign($1, $2)); }
;

/***********************************************************************
 * Helpers.
 ***********************************************************************/
id:
	  ID				{ $$ = U(SimpleName, $1); }
;

qid:
	  id
	| ID DOUBLECOLON ID		{ $$ = U(QualifiedName, P($1, $3)); }
;

vardef:
	type_exp id				{ $$ = make_vardef($1, $2); }
;

/***********************************************************************
 * Types.
 ***********************************************************************/
type_var_list_content:
	  TYVAR					{ $$ = CONS($1, nil); }
	| TYVAR ',' type_var_list_content	{ $$ = CONS($1, $3) }
;

type_var_list:
	  /* empty */				{ $$ = nil; }
	| '<' '>'				{ $$ = nil; }
	| '<' type_var_list_content '>'		{ $$ = $2; }
;

maybe_semicolons:
	  /* nothing */				{ $$ = nil; }
	| maybe_semicolons ';'			{ $$ = nil; }
;

record_type_fld:
	type_exp qid maybe_semicolons		{ $$ = make_field($1, $2); }
;

record_type_flds:
	  record_type_fld			{ $$ = CONS($1, nil); }
	| record_type_fld record_type_flds	{ $$ = CONS($1, $2); }
;

type_decl:
	  KW_STRUCT type_var_list ID '{' record_type_flds '}'
	  { $$ = make_tydef($3, $2, U(Struct, $5)); }
	| KW_OPT_STRUCT type_var_list ID '{' record_type_flds '}'
	  { $$ = make_tydef($3, $2, U(OptStruct, $5)); }
	| KW_UNION type_var_list ID '{' record_type_flds '}'
	  { $$ = make_tydef($3, $2, U(Union, $5)); }
	| KW_TYPEDEF type_exp type_var_list ID
	  { $$ = make_tydef($4, $3, $2); }
	| KW_TYPE type_var_list ID
	  { $$ = make_tydef($3, $2, U(AbstractType, NULL)); }
;

simple_type:
	  KW_VOID	{ $$ = U(Void, NULL); }
	| KW_INT	{ $$ = U(Int, NULL); }
	| KW_BOOL	{ $$ = U(Bool, NULL); }
	| KW_FLOAT	{ $$ = U(Float, NULL); }
	| KW_STRING	{ $$ = U(String, NULL); }
;

function_type:
	   '*' type_exp '(' func_args ')'
		{ $$ = bare_fn_type($2, $4); }
	 | '*' '(' type_exp '(' func_args ')' ')'
		{ $$ = bare_fn_type($3, $5); }
;

tuple_ty:
	  type_exp ',' ne_ty_list	{ $$ = CONS($1, $3); }
;

ne_ty_list:
	  type_exp			{ $$ = CONS($1, nil); }
	| type_exp ',' ne_ty_list	{ $$ = CONS($1, $3); }
;

ty_list:
	  /* empty */			{ $$ = nil; }
	| ne_ty_list			{ $$ = $1; }
;

xtype_exp:
	  simple_type			{ $$ = $1; }
	| function_type			{ $$ = $1; }
	| TYVAR				{ $$ = make_tyvar($1); }
	| '<' ty_list '>' qid		{ $$ = make_named_type($4, $2); }
	/* this is hack to allow <<t1>x>y */
	| LTLT ne_ty_list '>' qid '>' qid
	  { $$ = make_named_type($6, CONS(make_named_type($4, $2), nil)); }
	| LTLT ne_ty_list '>' qid ',' ne_ty_list '>' qid
	  { $$ = make_named_type($8, CONS(make_named_type($4, $2), $6)); }
	| '*' '[' tuple_ty ']'		{ $$ = U(Tuple, $3); }
;

type_exp:
	  '(' xtype_exp ')'		{ $$ = $2; }
	| qid				{ $$ = make_named_type($1, nil); }
	| xtype_exp			{ $$ = $1; }
;

/***********************************************************************
 * Prototypes.
 ***********************************************************************/
func_arg:
	  type_exp			{ $$ = P($1, none); }
	| type_exp ID			{ $$ = P($1, Some($2)); }
;

ne_func_args:
	  func_arg			{ $$ = CONS($1, nil); }
	| func_arg ',' ne_func_args	{ $$ = CONS($1, $3); }
;

func_args:
	  /* empty */			{ $$ = nil; }
	| ne_func_args			{ $$ = $1; }
;

func_header:
	type_exp id '(' func_args ')'
	{ $$ = make_vardef(bare_fn_type($1, $4), $2); }
;

func_proto:
	  func_header ';'		{ $$ = U(FuncProto, $1); }
	| func_header '=' string ';'	{ $$ = U(XCFuncProto, P($3, $1)); }
;

/***********************************************************************
 * Expressions.
 ***********************************************************************/
ne_expr_list:
	  expr				{ $$ = CONS($1, nil); }
	| expr ',' ne_expr_list		{ $$ = CONS($1, $3); }
;

expr_list:
	  /* empty */			{ $$ = nil; }
	| ne_expr_list			{ $$ = $1; }
;

variable:
	qid			{ $$ = EXPR(Ref, make_ref_record($1)); }
;

tuple_content:
	  expr ',' expr			{ $$ = CONS($1, CONS($3, nil)); }
	| expr ',' tuple_content	{ $$ = CONS($1, $3); }
;

tuple:
	'[' tuple_content ']'		{ $$ = EXPR(TupleExpr, $2); }
;

/* expression that might be used to form expression stmt */
stmt_expr:
	  call %prec APPLICATION	{ $$ = $1; }
	| expr '=' expr			{ $$ = BIN(Assign, $1, $3); }
	| MINUS_MINUS expr		{ $$ = UNARY(PreDec, $2); }
	| PLUS_PLUS expr		{ $$ = UNARY(PreInc, $2); }
	| expr MINUS_MINUS		{ $$ = UNARY(PostDec, $1); }
	| expr PLUS_PLUS		{ $$ = UNARY(PostInc, $1); }
;

pattern_expr:
	  primary			{ $$ = $1; }
	| pattern_expr '+' pattern_expr	{ $$ = BIN(Plus, $1, $3); }
	| pattern_expr '-' pattern_expr	{ $$ = BIN(Minus, $1, $3); }
	| pattern_expr '*' pattern_expr	{ $$ = BIN(Mult, $1, $3); }
	| pattern_expr '/' pattern_expr	{ $$ = BIN(Div, $1, $3); }
	| pattern_expr '%' pattern_expr	{ $$ = BIN(Mod, $1, $3); }
	| pattern_expr '<' pattern_expr	{ $$ = BIN(Lt, $1, $3); }
	| pattern_expr '>' pattern_expr	{ $$ = BIN(Gt, $1, $3); }
	| pattern_expr GE pattern_expr	{ $$ = BIN(Ge, $1, $3); }
	| pattern_expr LE pattern_expr	{ $$ = BIN(Le, $1, $3); }
	| pattern_expr EQ pattern_expr	{ $$ = BIN(Eq, $1, $3); }
	| pattern_expr NE pattern_expr	{ $$ = BIN(Ne, $1, $3); }
	| '-' pattern_expr %prec UMINUS { $$ = UNARY(Negate, $2); }
;

union_expr:
	/* union constructor */
	  qid '[' ']'
	  { $$ = EXPR(UnionExpr, make_namedexpr($1, EXPR(VoidConst, NULL))); }
	| qid '[' expr ']'
	  { $$ = EXPR(UnionExpr, make_namedexpr($1, $3)); }
	| qid tuple
	  { $$ = EXPR(UnionExpr, make_namedexpr($1, $2)); }
;

expr:
	  primary		{ $$ = $1; }
	/* TODO: we'll give slightly misleading error message because of
	 * this: */
	| '!' expr		{ $$ = BIN(Eq, $2, 
					   EXPR(BoolConst, (void*)0)); }
	| expr '+' expr		{ $$ = BIN(Plus, $1, $3); }
	| expr '-' expr		{ $$ = BIN(Minus, $1, $3); }
	| expr '*' expr		{ $$ = BIN(Mult, $1, $3); }
	| expr '/' expr		{ $$ = BIN(Div, $1, $3); }
	| expr '%' expr		{ $$ = BIN(Mod, $1, $3); }
	| expr '<' expr		{ $$ = BIN(Lt, $1, $3); }
	| expr '>' expr		{ $$ = BIN(Gt, $1, $3); }
	| expr GE expr		{ $$ = BIN(Ge, $1, $3); }
	| expr LE expr		{ $$ = BIN(Le, $1, $3); }
	| expr EQ expr		{ $$ = BIN(Eq, $1, $3); }
	| expr NE expr		{ $$ = BIN(Ne, $1, $3); }
	| expr AND_AND expr	{ $$ = BIN(AndIf, $1, $3); }
	| expr OR_OR expr	{ $$ = BIN(OrIf, $1, $3); }
	| '-' expr %prec UMINUS { $$ = UNARY(Negate, $2); }
	| tuple			{ $$ = $1; }
	| stmt_expr		{ $$ = $1; }
	| struct_expr		{ $$ = $1; }
	| union_expr		{ $$ = $1; }
;

primary:
	  INT
	  { $$ = make_int_const($1, make_location(&@$)) }
	| FLOAT
	  { $$ = make_float_const($1, make_location(&@$)) }
	| string		{ $$ = EXPR(StringConst, $1); }
	| KW_TRUE		{ $$ = EXPR(BoolConst, (void*)1); }
	| KW_FALSE		{ $$ = EXPR(BoolConst, (void*)0); }
	| variable		{ $$ = $1; }
	| '(' expr ')'		{ $$ = $2; }
	| lambda		{ $$ = $1; }
	| KW_NULL		{ $$ = EXPR(NullConst, NULL); }
	| primary '.' id
	  { $$ = EXPR(DotExpr, make_sini_expr($3, $1)); }
;

struct_init:
	qid '=' expr
	{ $$ = make_sini_expr($1, $3); }
;

struct_expr_content:
	  struct_init				{ $$ = CONS($1, nil); }
	| struct_init ',' struct_expr_content	{ $$ = CONS($1, $3); }
;

struct_expr:
	'(' '{' struct_expr_content '}' ')'
		{ $$ = EXPR(StructExpr, $3) }
;

call:
	primary '(' expr_list ')'
		{ $$ = EXPR(Call, P($1, $3)); }
;

lambda:
	  KW_FUN type_exp '(' func_args ')' fnc_body
		{ $$ = LOCATED(make_lambda(bare_fn_type($2, $4), $6)); }
;

/***********************************************************************
 * Statements.
 ***********************************************************************/
block:
	'{' stmts '}'		{ $$ = STMT(Block, $2); }
;

ne_stmts:
	  stmt	   		{ $$ = CONS($1, nil); }
	| stmt ne_stmts	   	{ $$ = CONS($1, $2); }
;

stmts:
	  /* empty */	   { $$ = nil; }
	| ne_stmts	   { $$ = $1; }
;

non_if_stmt:
	  normal_stmt ';'
	| block
	| func_def
	| loop_stmt
	| pattern_matching_stmt
	| try_stmt
;

no_else:	if_stmt | non_if_stmt;
stmt:		no_else | if_else_stmt;

if_stmt:
	KW_IF '(' expr ')' no_else
	{ $$ = STMT(If, gont_make_tuple(3, $3, $5, SKIP())); }
;

if_else_stmt:
	  KW_IF '(' expr ')' if_else_stmt KW_ELSE stmt
	  { $$ = STMT(If, gont_make_tuple(3, $3, $5, $7)); }
	| KW_IF '(' expr ')' non_if_stmt KW_ELSE stmt
	  { $$ = STMT(If, gont_make_tuple(3, $3, $5, $7)); }
;

single_var_def:
	  id			{ $$ = P($1, SKIP()); }
	| id '=' expr
	{ $$ = P($1, STMT(Expr, BIN(Assign,
				    EXPR(Ref, make_ref_record($1)),
				    $3))); }
	   				
;

var_def_list:
	  single_var_def			{ $$ = CONS($1, nil); }
	| single_var_def ',' var_def_list	{ $$ = CONS($1, $3); }
;

var_def:
	  type_exp var_def_list		{ $$ = make_vardefs($1, $2); }
;

/* Bison doesn't handle yyloc well if there are no tokes in lhs. */
for_exp1:
	  stmt_expr		{ $$ = STMT(Expr, $1); }
	| var_def		{ $$ = $1; }
	| KW_SKIP		{ $$ = SKIP(); }
	| /* empty */		
		{ $$ = make_located(U(Skip, NULL), &last_lex_loc); }
;

for_exp2:
	  expr			{ $$ = $1; }
	| /* empty */		
		{ $$ = make_expr(U(BoolConst, (void*)1), &last_lex_loc); }
;

for_exp3:
	  stmt_expr		{ $$ = STMT(Expr, $1) }
	| KW_SKIP		{ $$ = SKIP(); }
	| /* empty */		
		{ $$ = make_located(U(Skip, NULL), &last_lex_loc); }
;

maybe_label:
	/* empty */		{ $$ = empty_string; }
      | ID ':'			{ $$ = $1; }
;

maybe_id:
	/* empty */		{ $$ = U(SimpleName, empty_string); }
      | id			{ $$ = $1; }
;

loop_stmt:
	 maybe_label KW_WHILE '(' expr ')' stmt
	 { $$ = LOOP($1,
		     PAIR(STMT(BreakIfNot, $4), $6),
		     SKIP()); }
	| maybe_label KW_DO block KW_WHILE '(' expr ')' ';'
	  { $$ = LOOP($1,
		      PAIR($3, STMT(BreakIfNot, $6)),
		      SKIP()); }
	| maybe_label KW_FOR '('
		for_exp1 ';'
		for_exp2 ';'
		for_exp3 ')' stmt
	  { $$ = STMT(Block, CONS($4,
				  CONS(LOOP($1,
					    PAIR(STMT(BreakIfNot, $6), $10),
					    $8),
				       nil))); }
;

normal_stmt:
	  stmt_expr		{ $$ = STMT(Expr, $1); }
	| KW_RETURN		{ $$ = STMT(Return, EXPR(VoidConst, NULL)); }
	| KW_RETURN expr	{ $$ = STMT(Return, $2); }
	| KW_SKIP		{ $$ = SKIP(); }
	| var_def		{ $$ = $1; }
	| KW_RAISE expr		{ $$ = STMT(Raise, $2); }
	| KW_BREAK maybe_id
		{ $$ = LOCATED(make_break(U(BreakKind, NULL), $2)); }
	| KW_CONTINUE maybe_id
		{ $$ = LOCATED(make_break(U(ContinueKind, NULL), $2)); }
;

fnc_body:
	  block		{ $$ = $1; }
	| '(' expr ')'	{ $$ = STMT(Block, CONS(STMT(Return, $2), nil)); }
;

/***********************************************************************
 * Pattern matching statements.
 ***********************************************************************/

one_case:
	  KW_CASE pattern ':'	
	  	{ $$ = P($2, EXPR(BoolConst, (void*)1)); }
	| KW_CASE pattern KW_IF expr ':'	
		{ $$ = P($2, $4); }
;

cases:
	  one_case		{ $$ = CONS($1, nil); }
	| one_case cases	{ $$ = CONS($1, $2); }
;

case_clause:
	cases ne_stmts		{ $$ = make_case_clause($1, STMT(Block, $2)); }
;

case_list:
	  case_clause case_list		{ $$ = CONS($1, $2); }
	| /* empty */			{ $$ = nil; }
;

pattern_eq:
	  pattern '=' expr			{ $$ = P($1, $3); }
;

pattern_eq_list:
	  pattern_eq				{ $$ = CONS($1, nil); }
	| pattern_eq ',' pattern_eq_list	{ $$ = CONS($1, $3); }
;

maybe_finally:
	  KW_FINALLY stmt			{ $$ = $2; }
	| /* nothing */				{ $$ = SKIP(); }
;

maybe_with:
	  KW_WITH '{' case_list '}'		{ $$ = $3; }
	| /* nothing */				{ $$ = nil; }
;

try_stmt:
	KW_TRY block maybe_with maybe_finally
			{ $$ = STMT(Try, make_try_record($2, $3, $4)); }
;

pattern_matching_stmt:
	  KW_LET pattern_eq_list KW_IN stmt
	  { $$ = STMT(Let, P($2, $4)); }
	| KW_SWITCH expr '{' case_list '}'
	  { $$ = STMT(Switch, P($2, $4)); }
;


/***********************************************************************
 * Patterns.
 ***********************************************************************/
ne_pattern_list:
	  pattern ',' ne_pattern_list	{ $$ = CONS($1, $3); }
	| pattern			{ $$ = CONS($1, nil); }
;

pattern_list:
	  /* empty */			{ $$ = nil; }
	| ne_pattern_list		{ $$ = $1; }
;

raw_tuple_pattern:
	'[' pattern_list ']'		{ $$ = U(TuplePattern, $2); }
;

tuple_pattern:
	raw_tuple_pattern		{ $$ = LOCATED($1); }
;

raw_pattern:
	  '_'				{ $$ = U(VoidPattern, NULL); }
	| pattern_expr			{ $$ = U(ExprPattern, $1); }
	| raw_tuple_pattern		{ $$ = $1; }
	| qid tuple_pattern
	  { $$ = U(UnionPattern, make_named_pattern($1, $2)); }
;

pattern:
	raw_pattern			{ $$ = LOCATED($1); }
;

%%

struct gont_union *parse(struct gont_string *file)
{
	void *lex = make_lexer(file->data);

	/* yydebug = 1; */

	if (none == NULL) {
		none = gont_make_union(GontUF_Std__None, NULL);
		nil = gont_make_union(GontUF_List__Nil, NULL);
		empty_string = gont_make_string(0, NULL);
	}

	input_filename = gont_make_string(file->len, file->data);

	if (yyparse(lex)) {
		/* We should never get here... */
		yyerror("parse error");
	}

	return decls;
}
