/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * Util - Various utility functions: diagnostics, compilation flags, etc.
 */
 
open Ast;

/* Prints error message, I guess not really usefull. */
void message(location loc, string msg);

/* Prints message prepended with 'warning: ', sets bailout flag is -Werror
 * was given. */
void warnl(location loc, string msg);

/* Prints message and sets bailout flag. */
void errorl(location loc, string msg);

/* If bailout flag is set raise Exit. */
void maybe_choke();

/* Location stack stuff: */
void push_location(location loc);
void pop_location();
location top_location();

/* Use top_location(): */
void warn(string msg);
void error(string msg);

/* Run given function on x.a with location set to x.loc. */
'a locate(*('a ('b)) f, <'b>located x);

/* Prints error message and does abort(). */
'a ice(string msg);

/* If symbol is not bound -- ICE. */
'a get_binding(<'a>bindable sym);

/* Flags */
struct compiler_flags {
	bool werror;
	bool wshadow;
	bool wdead;
	bool wlabeled;
	bool wpattern;
	bool wcapitalization;
	bool verbose_errors;
	bool time_report;
}

compiler_flags flags;

/* Same as print_string(). */
void dbg(string);


*('a ('b)) partial(*('a ('c, 'b)) f, 'c c);
*('a ('b)) partial2(*('a ('c, 'd, 'b)) f, 'c c, 'd d);

// Return true is two given symbols are the same.
bool same_symbols(<'a>symbol s1, <'a>symbol s2);

// typing stuff
ty type_of(expr e);
ft_record get_ft_record(ty t);

// return uniqe name starting with [prefix]
string genname(string prefix);

// return string being flat terse representation of given type, using 
// [A-Za-z_0-9:<>] characters only
string mangle_type(ty t);
