/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * Xobjects - .xg{o,a} save/load functions.
 */
 
open Ast;
open Xobjects;

Xml::node serialize_xgo(xgo x)
{
    Xml::node named(string elname, string objname)
    {
	return Xml::Element[elname, List::Nil, List::cons1(["name", objname])];
    }

    Xml::node list(string pname, string ename, <id>List::t lst)
    {
	Xml::node one_element(id i)
	{
	    return named(ename, i);
	}
	
	let elements = List::map(one_element, lst) in
	return Xml::Element[pname, elements, List::Nil];
    }

    let refs = list("references", "module", x.xgo_references) in
    let uses = list("uses", "module", x.xgo_uses) in
    let exns = list("used-exns", "exception", x.xgo_used_exns) in 
    let ctor = named("ctor-routine", x.xgo_ctor_routine) in
    let init = named("init-routine", x.xgo_init_routine) in
    let fini = named("fini-routine", x.xgo_fini_routine) in {
	<Xml::node>List::t children = List::cons3(refs, uses, exns);

	children = List::Cons[ctor, List::Cons[init, List::Cons[fini, children]]];
    
	return Xml::Element["object", children, 
			    List::cons1(["name", x.xgo_name])];
    }
}

id get_name(Xml::node n)
{
    return Xml::prop_value("name", n);
}
    
xgo unserialize_xgo(Xml::node n)
{
    <id>List::t list(id pname, id elname)
    {
	id one(Xml::node e)
	{
	    switch e {
	    case Xml::Element[nam, List::Nil, 
			      List::Cons[["name", i], List::Nil]]:
		if (nam == elname)
		    return i;
	    case _:
		skip;
	    }

	    raise Not_found;
	}
	
	switch Xml::find_element(pname, n) {
	case Xml::Element[_, children, _]:
	    return List::map(one, children);
	    
	case _:
	    return abort();
	}
    }

    try {
	xgo x = ({
	    xgo_references = list("references", "module"),
	    xgo_uses = list("uses", "module"),
	    xgo_used_exns = list("used-exns", "exception"),
	    xgo_name = get_name(n),
	    xgo_ctor_routine = get_name(Xml::find_element("ctor-routine", n)),
	    xgo_init_routine = get_name(Xml::find_element("init-routine", n)),
	    xgo_fini_routine = get_name(Xml::find_element("fini-routine", n)),
	    xgo_weak = false
	});

	return x;
    } with {
    case Not_found:
	raise Invalid_argument["Linker::unserialize_xgo"];
    }
}

Xml::node serialize_xga(xga x)
{
    let mods = List::map(serialize_xgo, x.xga_modules) in
    let mods_node = Xml::Element["modules", mods, List::Nil] in
    let link_flags = Xml::Element["link-flags", List::Nil, 
				  List::cons1(["value", x.xga_link_flags])] in
    return Xml::Element["library", List::cons2(link_flags, mods_node),
		        List::cons1(["name", x.xga_name])];
}

xga unserialize_xga(Xml::node n)
{
    try {
	<xgo>List::t mods;

	switch Xml::find_element("modules", n) {
	case Xml::Element[_, ch, _]:
	    mods = List::map(unserialize_xgo, ch);
	case _:
	    return abort();
	}
	
	let link_flags = Xml::prop_value("value", 
					 Xml::find_element("link-flags", n)) in
	return ({
	    xga_name = get_name(n),
	    xga_link_flags = link_flags,
	    xga_modules = mods
	});

    } with {
    case Not_found:
	raise Invalid_argument["Linker::unserialize_xga"];
    }
}

void save_xgo(xgo x, string filename)
{
    Xml::save_to_file_indented(Xml::make_doc(serialize_xgo(x)), filename);
}

void save_xga(xga x, string filename)
{
    Xml::save_to_file_indented(Xml::make_doc(serialize_xga(x)), filename);
}

// vim:sts=4:sw=4
